@@ -2,6 +2,7 @@
     ksym.c - functions for kernel address->symbol translation
     Copyright (c) 1995, 1996  Dr. G.W. Wettstein <greg@wind.rmcc.com>
     Copyright (c) 1996 Enjellic Systems Development
+    Copyright (c) 1997-2007 Martin Schulze <joey@infodrom.org>
 
     This file is part of the sysklogd package, a kernel and system log daemon.
 
@@ -105,6 +106,23 @@
  *
  * Tue Sep 12 23:48:12 CEST 2000: Martin Schulze <joey@infodrom.ffis.de>
  *	Close symbol file in InitKsyms() when an error occurred.
+ *
+ * Thu Apr 29 18:07:16 CEST 2004: Dmitry Levin <ldv@altlinux.org>
+ *	Close file descriptor in FindSymbolFile() in order not to leak
+ *	file descriptors.
+ *
+ * Fri Jul 16 08:32:49 CEST 2004: Ulf Hrnhammar <Ulf.Harnhammar.9485@student.uu.se>
+ *	Added boundary check for fscanf() in InitKsyms() and
+ *	CheckMapVersion() to prevent an unintended crash when reading
+ *	an incorrect System.map.
+ *
+ * Mon May 28 08:27:51 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Added back /usr/src/linux/System.map as fall-back location.
+ *
+ * Thu May 31 16:56:26 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Improved symbol lookup, since symbols are spread over the entire
+ *	address space.  Return the symbol that fits best instead of
+ *	the first hit.
  */
 
 
@@ -114,18 +132,12 @@
 #include <sys/utsname.h>
 #include "klogd.h"
 #include "ksyms.h"
+#include "module.h"
 
 #define VERBOSE_DEBUGGING 0
 
 
-/* Variables static to this module. */
-struct sym_table
-{
-	unsigned long value;
-	char *name;
-};
-
-static int num_syms = 0;
+int num_syms = 0;
 static int i_am_paranoid = 0;
 static char vstring[12];
 static struct sym_table *sym_array = (struct sym_table *) 0;
@@ -134,6 +146,7 @@ static char *system_maps[] =
 {
 	"/boot/System.map",
 	"/System.map",
+	"/usr/src/linux/System.map",
 #if defined(TEST)
 	"./System.map",
 #endif
@@ -162,7 +175,7 @@ static int CheckMapVersion(char *);
  * Purpose:	This function is responsible for initializing and loading
  *		the data tables used by the kernel address translations.
  *
- * Arguements:	(char *) mapfile
+ * Arguments:	(char *) mapfile
  *
  *			mapfile:->	A pointer to a complete path
  *					specification of the file containing
@@ -236,7 +249,7 @@ extern int InitKsyms(mapfile)
 	 */
 	while ( !feof(sym_file) )
 	{
-		if ( fscanf(sym_file, "%lx %c %s\n", &address, &type, sym)
+		if ( fscanf(sym_file, "%lx %c %511s\n", &address, &type, sym)
 		    != 3 )
 		{
 			Syslog(LOG_ERR, "Error in symbol table input (#1).");
@@ -303,7 +316,7 @@ extern int InitKsyms(mapfile)
  *		in /boot will be skipped and the map in /usr/src/linux would
  *		be used if its version number matches the executing kernel.
  *
- * Arguements:	None specified.
+ * Arguments:	None specified.
  *
  * Return:	char *
  *
@@ -344,6 +357,7 @@ static char * FindSymbolFile()
 		if ( (sym_file = fopen(symfile, "r")) != (FILE *) 0 ) {
 			if (CheckMapVersion(symfile) == 1)
 				file = symfile;
+			fclose (sym_file);
 		}
 		if (sym_file == (FILE *) 0 || file == (char *) 0) {
 			sprintf (symfile, "%s", *mf);
@@ -352,6 +366,7 @@ static char * FindSymbolFile()
 			if ( (sym_file = fopen(symfile, "r")) != (FILE *) 0 ) {
 				if (CheckMapVersion(symfile) == 1)
 					file = symfile;
+				fclose (sym_file);
 			}
 		}
 
@@ -387,7 +402,7 @@ static char * FindSymbolFile()
  *		needs more than 255 patch levels to get a kernel out the
  *		door... :-)
  *
- * Arguements:	(char *) version
+ * Arguments:	(char *) version
  *
  *			version:->	A pointer to the string which
  *					is to be decoded as a kernel
@@ -493,7 +508,7 @@ static int CheckVersion(version)
  *		currently running kernel.  It uses CheckVersion as
  *		backend.
  *
- * Arguements:	(char *) fname
+ * Arguments:	(char *) fname
  *
  *			fname:->	A pointer to the string which
  *					references the system map file to
@@ -533,7 +548,7 @@ static int CheckMapVersion(fname)
 		version = 0;
 		while ( !feof(sym_file) && (version == 0) )
 		{
-			if ( fscanf(sym_file, "%lx %c %s\n", &address, \
+			if ( fscanf(sym_file, "%lx %c %511s\n", &address, \
 				    &type, sym) != 3 )
 			{
 				Syslog(LOG_ERR, "Error in symbol table input (#2).");
@@ -580,7 +595,7 @@ static int CheckMapVersion(fname)
  * Purpose:	This function is responsible for adding a symbol name
  *		and its address to the symbol table.
  *
- * Arguements:	(unsigned long) address, (char *) symbol
+ * Arguments:	(unsigned long) address, (char *) symbol
  *
  * Return:	int
  *
@@ -620,7 +635,7 @@ static int AddSymbol(address, symbol)
  * Purpose:	Find the symbol which is related to the given kernel
  *		address.
  *
- * Arguements:	(long int) value, (struct symbol *) sym
+ * Arguments:	(long int) value, (struct symbol *) sym
  *
  *		value:->	The address to be located.
  * 
@@ -644,30 +659,52 @@ char * LookupSymbol(value, sym)
 	auto int lp;
 	
 	auto char *last;
+	auto char *name;
+
+	struct symbol ksym, msym;
 
 	if (!sym_array)
 		return((char *) 0);
 
 	last = sym_array[0].name;
-	sym->offset = 0;
-	sym->size = 0;
+	ksym.offset = 0;
+	ksym.size = 0;
 	if ( value < sym_array[0].value )
 		return((char *) 0);
 	
-	for(lp= 0; lp <= num_syms; ++lp)
+	for(lp = 0; lp <= num_syms; ++lp)
 	{
 		if ( sym_array[lp].value > value )
 		{		
-			sym->offset = value - sym_array[lp-1].value;
-			sym->size = sym_array[lp].value - \
+			ksym.offset = value - sym_array[lp-1].value;
+			ksym.size = sym_array[lp].value - \
 				sym_array[lp-1].value;
-			return(last);
+			break;
 		}
 		last = sym_array[lp].name;
 	}
 
-	if ( (last = LookupModuleSymbol(value, sym)) != (char *) 0 )
+	name = LookupModuleSymbol(value, &msym);
+
+	if ( ksym.offset == 0 && msym.offset == 0 )
+	{
+		return((char *) 0);
+	}
+
+	if ( ksym.offset == 0 || msym.offset < 0 ||
+	     (ksym.offset > 0 && ksym.offset < msym.offset) )
+	{
+		sym->offset = ksym.offset;
+		sym->size = ksym.size;
 		return(last);
+	}
+	else
+	{
+		sym->offset = msym.offset;
+		sym->size = msym.size;
+		return(name);
+	}
+
 
 	return((char *) 0);
 }
@@ -681,7 +718,7 @@ char * LookupSymbol(value, sym)
  *		also initializes the symbol count and in general prepares
  *		for a re-read of a static symbol table.
  *
- * Arguements:  void
+ * Arguments:  void
  *
  * Return:	void
  **************************************************************************/
@@ -692,7 +729,7 @@ static void FreeSymbols()
 	auto int lp;
 
 	/* Free each piece of memory allocated for symbol names. */
-	for(lp= 0; lp < num_syms; ++lp)
+	for(lp = 0; lp < num_syms; ++lp)
 		free(sym_array[lp].name);
 
 	/* Whack the entire array and initialize everything. */
@@ -711,7 +748,7 @@ static void FreeSymbols()
  *		line after all potential numeric kernel addresses have
  *		been resolved symolically.
  *
- * Arguements:	(char *) line, (char *) el
+ * Arguments:	(char *) line, (char *) el
  *
  *		line:->	A pointer to the buffer containing the kernel
  *			message to be expanded and logged.
@@ -740,6 +777,8 @@ extern char * ExpandKadds(line, el)
 
 	auto struct symbol sym;
 
+	sym.offset = 0;
+	sym.size = 0;
 
 	/*
 	 * This is as handy a place to put this as anyplace.
@@ -806,7 +845,7 @@ extern char * ExpandKadds(line, el)
 		{
 			--value;
 			++kp;
-			elp += sprintf(elp, "+%x/%d", sym.offset, sym.size);
+			elp += sprintf(elp, "+0x%x/0x%02x", sym.offset, sym.size);
 		}
 		strncat(elp, kp, value);
 		elp += value;
@@ -829,7 +868,7 @@ extern char * ExpandKadds(line, el)
  *		mode of loadable module symbol lookups.  Probably overkill
  *		but it does slay another global variable.
  *
- * Arguements:	(int) level
+ * Arguments:	(int) level
  *
  *		level:->	The amount of paranoia which is to be
  *				present when resolving kernel exceptions.
@@ -899,3 +938,11 @@ extern void Syslog(int priority, char *fmt, ...)
 	return;
 }
 #endif
+
+/*
+ * Local variables:
+ *  c-indent-level: 8
+ *  c-basic-offset: 8
+ *  tab-width: 8
+ * End:
+ */
