@@ -417,12 +417,6 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  *	file is defined in the used libc and should not be hardcoded
  *	into the syslogd binary referring the system it was compiled on.
  *
- * Sun Sep 17 20:45:33 CEST 2000: Martin Schulze <joey@infodrom.ffis.de>
- *	Fixed some bugs in printline() code that did not escape
- *	control characters '\177' through '\237' and contained a
- *	single-byte buffer overflow.  Thanks to Solar Designer
- *	<solar@false.com>.
- *
  * Sun Sep 17 21:26:16 CEST 2000: Martin Schulze <joey@infodrom.ffis.de>
  *	Don't close open sockets upon reload.  Thanks to Bill
  *	Nottingham.
@@ -441,6 +435,65 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  *	Don't return a closed fd if `-a' is called with a wrong path.
  *	Thanks to Bill Nottingham <notting@redhat.com> for providing
  *	a patch.
+ * Thu Apr 13 05:08:10 CEST 2001: Jon Burgess <Jon_Burgess@eur.3com.com>
+ *	Moved the installation of the signal handler up a little bit
+ *	so it guaranteed to be available when the child is forked,
+ *	hence, fixing a  race condition.  This used to create problems
+ *	with UML and fast machines.
+ *
+ * Sat Apr 17 18:03:05 CEST 2004: Steve Grubb <linux_4ever@yahoo.com>
+ *	Correct memory allocation for for commandline arguments in
+ *	crunch_list().
+ *
+ * Thu Apr 29 12:38:39 CEST 2004: Solar Designer <solar@openwall.com>
+ *	Applied Openwall paranoia patches to improve crunch_list().
+ *
+ * Tue May  4 16:47:30 CEST 2004: Solar Designer <solar@openwall.com>
+ *	Ensure that "len" is not placed in a register, and that the
+ *	endtty() signal handler is not installed too early which could
+ *	cause a segmentation fault or worse.
+ *
+ * Tue May  4 16:52:01 CEST 2004: Solar Designer <solar@openwall.com>
+ *	Adjust the size of a variable to prevent a buffer overflow
+ *	should _PATH_DEV ever contain something different than "/dev/".
+ *
+ * Tue Nov  2 20:28:23 CET 2004: Colin Phipps <cph@cph.demon.co.uk>
+ *	Don't block on the network socket, in case a packet gets lost
+ *	between select and recv.
+ *
+ * Sun Nov  7 12:28:47 CET 2004: Martin Schulze <joey@infodrom.org>
+ *	Discard any timestamp information found in received syslog
+ *	messages.  This will affect local messages sent from a
+ *	different timezone.
+ *
+ * Sun Nov  7 13:47:00 CET 2004: Martin Schulze <joey@infodrom.org>
+ *	Remove trailing newline when forwarding messages.
+ *
+ * Thu May 25 09:47:38 CEST 2006: Martin Schulze <joey@infodrom.org>
+ *	Reset the 'restart' flag immediately after entering the
+ *	restart code, so that subsequent SIGHUPs are able to set it
+ *	again and cause a new restart.  This fixes a race condition
+ *	when somebody sends tons of HUP signals.
+ *
+ * Thu May 24 15:24:49 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Ignore errors caused by filled up disks so that the log
+ *	continues to be written as soon as space becomes available
+ *	again.
+ *
+ * Sat May 26 10:05:05 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Only try to gather the local domain name when messages are to
+ *	be received from the network, it's not needed otherwise.
+ *
+ * Sat May 26 12:22:44 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Properly accompany the MARK message with the facility.
+ *
+ * Mon May 28 19:44:39 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Notify the waiting parent process if the client dies to it
+ *	doesn't wait the entire five minutes.
+ *
+ * Wed Jul  4 21:02:22 CEST 2007: Martin Schulze <joey@infodrom.org>
+ *	Open a pipe with O_NOCTTY to avoid them becoming the controlling
+ *	tty and normal files with O_NONBLOCK to avoid blocking.
  */
 
 
@@ -723,6 +776,7 @@ struct code	FacNames[] = {
 int	Debug;			/* debug flag */
 char	LocalHostName[MAXHOSTNAMELEN+1];	/* our hostname */
 char	*LocalDomain;		/* our local domain name */
+char	*emptystring = "";
 int	InetInuse = 0;		/* non-zero if INET sockets are being used */
 int	finet = -1;		/* Internet datagram socket */
 int	LogPort;		/* port number for INET connections */
@@ -782,11 +836,14 @@ int main(argc, argv)
 {
 	register int i;
 	register char *p;
+#ifndef TESTING
+	ssize_t msglen;
+#endif
 #if !defined(__GLIBC__)
 	int len, num_fds;
 #else /* __GLIBC__ */
 #ifndef TESTING
-	size_t len;
+	socklen_t len;
 #endif
 	int num_fds;
 #endif /* __GLIBC__ */
@@ -890,11 +947,11 @@ int main(argc, argv)
 		dprintf("Checking pidfile.\n");
 		if (!check_pid(PidFile))
 		{
+			signal (SIGTERM, doexit);
 			if (fork()) {
 				/*
 				 * Parent process
 				 */
-				signal (SIGTERM, doexit);
 				sleep(300);
 				/*
 				 * Not reached unless something major went wrong.  5
@@ -906,6 +963,7 @@ int main(argc, argv)
 				 */
 				exit(1);
 			}
+			signal (SIGTERM, SIG_DFL);
 			num_fds = getdtablesize();
 			for (i= 0; i < num_fds; i++)
 				(void) close(i);
@@ -935,12 +993,16 @@ int main(argc, argv)
 			if (!write_pid(PidFile))
 			{
 				dprintf("Can't write pid.\n");
+				if (getpid() != ppid)
+					kill (ppid, SIGTERM);
 				exit(1);
 			}
 		}
 		else
 		{
 			dprintf("Pidfile (and pid) already exist.\n");
+			if (getpid() != ppid)
+				kill (ppid, SIGTERM);
 			exit(1);
 		}
 	} /* if ( !Debug ) */
@@ -949,14 +1011,13 @@ int main(argc, argv)
 	consfile.f_type = F_CONSOLE;
 	(void) strcpy(consfile.f_un.f_fname, ctty);
 	(void) gethostname(LocalHostName, sizeof(LocalHostName));
+	LocalDomain = emptystring;
 	if ( (p = strchr(LocalHostName, '.')) ) {
 		*p++ = '\0';
 		LocalDomain = p;
 	}
-	else
+	else if ( AcceptRemote )
 	{
-		LocalDomain = "";
-
 		/*
 		 * It's not clearly defined whether gethostname()
 		 * should return the simple hostname or the fqdn. A
@@ -992,6 +1053,7 @@ int main(argc, argv)
 	(void) signal(SIGCHLD, reapchild);
 	(void) signal(SIGALRM, domark);
 	(void) signal(SIGUSR1, Debug ? debug_switch : SIG_IGN);
+	(void) signal(SIGXFSZ, SIG_IGN);
 	(void) alarm(TIMERINTVL);
 
 	/* Create a partial message table for all file descriptors. */
@@ -1001,6 +1063,10 @@ int main(argc, argv)
 	    (char **) 0 )
 	{
 		logerror("Cannot allocate memory for message parts table.");
+#ifndef TESTING
+		if (getpid() != ppid)
+			kill (ppid, SIGTERM);
+#endif
 		die(0);
 	}
 	for(i= 0; i < num_fds; ++i)
@@ -1074,9 +1140,9 @@ int main(argc, argv)
 				  (fd_set *) NULL, (struct timeval *) NULL);
 		if ( restart )
 		{
+			restart = 0;
 			dprintf("\nReceived SIGHUP, reloading syslogd.\n");
 			init();
-			restart = 0;
 			continue;
 		}
 		if (nfds == 0) {
@@ -1104,31 +1170,29 @@ int main(argc, argv)
 #ifdef SYSLOG_UNIXAF
 		for (i = 0; i < nfunix; i++) {
 		    if ((fd = funix[i]) != -1 && FD_ISSET(fd, &readfds)) {
-			memset(line, '\0', sizeof(line));
-			i = recv(fd, line, MAXLINE - 2, 0);
+			memset(line, 0, sizeof(line));
+			msglen = recv(fd, line, MAXLINE - 2, 0);
 			dprintf("Message from UNIX socket: #%d\n", fd);
-			if (i > 0) {
-				line[i] = line[i+1] = '\0';
-				printchopped(LocalHostName, line, i + 2,  fd);
-			} else if (i < 0 && errno != EINTR) {
+			if (msglen > 0)
+				printchopped(LocalHostName, line, msglen + 2,  fd);
+			else if (msglen < 0 && errno != EINTR) {
 				dprintf("UNIX socket error: %d = %s.\n", \
 					errno, strerror(errno));
 				logerror("recvfrom UNIX");
-	      	}
-				}
 			}
+		    }
+		}
 #endif
 
 #ifdef SYSLOG_INET
 		if (InetInuse && AcceptRemote && FD_ISSET(inetm, &readfds)) {
 			len = sizeof(frominet);
-			memset(line, '\0', sizeof(line));
-			i = recvfrom(finet, line, MAXLINE - 2, 0, \
+			memset(line, 0, sizeof(line));
+			msglen = recvfrom(finet, line, MAXLINE - 2, 0, \
 				     (struct sockaddr *) &frominet, &len);
 			dprintf("Message from inetd socket: #%d, host: %s\n",
 				inetm, inet_ntoa(frominet.sin_addr));
-			if (i > 0) {
-				line[i] = line[i+1] = '\0';
+			if (msglen > 0) {
 				from = (char *)cvthname(&frominet);
 				/*
 				 * Here we could check if the host is permitted
@@ -1140,14 +1204,14 @@ int main(argc, argv)
 				 *  -Joey
 				 */
 				printchopped(from, line, \
- 					     i + 2,  finet);
-			} else if (i < 0 && errno != EINTR) {
+ 					     msglen + 2,  finet);
+			} else if (msglen < 0 && errno != EINTR && errno != EAGAIN) {
 				dprintf("INET socket error: %d = %s.\n", \
 					errno, strerror(errno));
 				logerror("recvfrom inet");
 				/* should be harmless now that we set
 				 * BSDCOMPAT on the socket */
-				sleep(10);
+				sleep(1);
 			}
 		}
 #endif
@@ -1216,6 +1280,7 @@ static int create_inet_socket()
 {
 	int fd, on = 1;
 	struct sockaddr_in sin;
+	int sockflags;
 
 	fd = socket(AF_INET, SOCK_DGRAM, 0);
 	if (fd < 0) {
@@ -1232,12 +1297,21 @@ static int create_inet_socket()
 		close(fd);
 		return -1;
 	}
-	/* We need to enable BSD compatibility. Otherwise an attacker
-	 * could flood our log files by sending us tons of ICMP errors.
+	/* We must not block on the network socket, in case a packet
+	 * gets lost between select and recv, otherise the process
+	 * will stall until the timeout, and other processes trying to
+	 * log will also stall.
 	 */
-	if (setsockopt(fd, SOL_SOCKET, SO_BSDCOMPAT, \
-			(char *) &on, sizeof(on)) < 0) {
-		logerror("setsockopt(BSDCOMPAT), suspending inet");
+	if ((sockflags = fcntl(fd, F_GETFL)) != -1) {
+		sockflags |= O_NONBLOCK;
+		/*
+		 * SETFL could fail too, so get it caught by the subsequent
+		 * error check.
+		 */
+		sockflags = fcntl(fd, F_SETFL, sockflags);
+	}
+	if (sockflags == -1) {
+		logerror("fcntl(O_NONBLOCK), suspending inet");
 		close(fd);
 		return -1;
 	}
@@ -1254,30 +1328,26 @@ char **
 crunch_list(list)
 	char *list;
 {
-	int count, i;
+	int i, m, n;
 	char *p, *q;
 	char **result = NULL;
 
 	p = list;
 	
 	/* strip off trailing delimiters */
-	while (p[strlen(p)-1] == LIST_DELIMITER) {
-		count--;
+	while (*p && p[strlen(p)-1] == LIST_DELIMITER)
 		p[strlen(p)-1] = '\0';
-	}
 	/* cut off leading delimiters */
-	while (p[0] == LIST_DELIMITER) {
-		count--;
+	while (p[0] == LIST_DELIMITER)
 		p++; 
-	}
 	
-	/* count delimiters to calculate elements */
-	for (count=i=0; p[i]; i++)
-		if (p[i] == LIST_DELIMITER) count++;
+	/* count delimiters to calculate the number of elements */
+	for (n = i = 0; p[i]; i++)
+		if (p[i] == LIST_DELIMITER) n++;
 	
-	if ((result = (char **)malloc(sizeof(char *) * count+2)) == NULL) {
+	if ((result = (char **)malloc(sizeof(char *) * (n + 2))) == NULL) {
 		printf ("Sorry, can't get enough memory, exiting.\n");
-		exit(0);
+		exit(1);
 	}
 	
 	/*
@@ -1285,30 +1355,28 @@ crunch_list(list)
 	 * characters are different from any delimiters,
 	 * so we don't have to care about this.
 	 */
-	count = 0;
-	while ((q=strchr(p, LIST_DELIMITER))) {
-		result[count] = (char *) malloc((q - p + 1) * sizeof(char));
-		if (result[count] == NULL) {
+	m = 0;
+	while ((q = strchr(p, LIST_DELIMITER)) && m < n) {
+		result[m] = (char *) malloc((q - p + 1) * sizeof(char));
+		if (result[m] == NULL) {
 			printf ("Sorry, can't get enough memory, exiting.\n");
-			exit(0);
+			exit(1);
 		}
-		strncpy(result[count], p, q - p);
-		result[count][q - p] = '\0';
+		memcpy(result[m], p, q - p);
+		result[m][q - p] = '\0';
 		p = q; p++;
-		count++;
+		m++;
 	}
-	if ((result[count] = \
-	     (char *)malloc(sizeof(char) * strlen(p) + 1)) == NULL) {
+	if ((result[m] = strdup(p)) == NULL) {
 		printf ("Sorry, can't get enough memory, exiting.\n");
-		exit(0);
+		exit(1);
 	}
-	strcpy(result[count],p);
-	result[++count] = NULL;
+	result[++m] = NULL;
 
 #if 0
-	count=0;
-	while (result[count])
-		dprintf ("#%d: %s\n", count, StripDomains[count++]);
+	m = 0;
+	while (result[m])
+		dprintf ("#%d: %s\n", m, result[m++]);
 #endif
 	return result;
 }
@@ -1444,16 +1512,11 @@ void printline(hname, msg)
 	memset (line, 0, sizeof(line));
 	q = line;
 	while ((c = *p++) && q < &line[sizeof(line) - 4]) {
-		if (c == '\n')
+		if (c == '\n' || c == 127)
 			*q++ = ' ';
 		else if (c < 040) {
 			*q++ = '^';
 			*q++ = c ^ 0100;
-		} else if (c == 0177 || (c & 0177) < 040) {
-			*q++ = '\\';
-			*q++ = '0' + ((c & 0300) >> 6);
-			*q++ = '0' + ((c & 0070) >> 3);
-			*q++ = '0' + (c & 0007);
 		} else
 			*q++ = c;
 	}
@@ -1539,35 +1602,50 @@ void logmsg(pri, msg, from, flags)
 	int fac, prilev, lognum;
 	int msglen;
 	char *timestamp;
+#ifdef __gnu_linux__
+	sigset_t mask;
+#else
+#ifndef SYSV
+	sigset_t omask;
+#endif
+#endif
 
 	dprintf("logmsg: %s, flags %x, from %s, msg %s\n", textpri(pri), flags, from, msg);
 
+#ifdef __gnu_linux__
+	sigemptyset(&mask);
+	sigaddset(&mask, SIGHUP);
+	sigaddset(&mask, SIGALRM);
+	sigprocmask(SIG_BLOCK, &mask, NULL);
+#else
 #ifndef SYSV
 	omask = sigblock(sigmask(SIGHUP)|sigmask(SIGALRM));
+#endif
 #endif
 
 	/*
 	 * Check to see if msg looks non-standard.
+	 *
+	 * A message looks like
+	 * Nov 17 11:42:33 CRON[
+	 * 01234567890123456
+	 *    ^  ^  ^  ^  ^
+	 *
+	 * Remote messages are not accompanied by a timestamp.
+	 * Local messages are accompanied by a timestamp (program's timezone)
 	 */
 	msglen = strlen(msg);
-	if (msglen < 16 || msg[3] != ' ' || msg[6] != ' ' ||
-	    msg[9] != ':' || msg[12] != ':' || msg[15] != ' ')
-		flags |= ADDDATE;
-
-	(void) time(&now);
-	if (flags & ADDDATE)
-		timestamp = ctime(&now) + 4;
-	else {
-		timestamp = msg;
+	if (!(msglen < 16 || msg[3] != ' ' || msg[6] != ' ' ||
+	    msg[9] != ':' || msg[12] != ':' || msg[15] != ' ')) {
 		msg += 16;
 		msglen -= 16;
 	}
 
+	(void) time(&now);
+	timestamp = ctime(&now) + 4;
+
 	/* extract facility and priority level */
-	if (flags & MARK)
-		fac = LOG_NFACILITIES;
-	else
-		fac = LOG_FAC(pri);
+	fac = LOG_FAC(pri);
 	prilev = LOG_PRI(pri);
 
 	/* log the message to the particular outputs */
@@ -1581,8 +1659,12 @@ void logmsg(pri, msg, from, flags)
 			(void) close(f->f_file);
 			f->f_file = -1;
 		}
+#ifdef __gnu_linux__
+		sigprocmask(SIG_UNBLOCK, &mask, NULL);
+#else
 #ifndef SYSV
 		(void) sigsetmask(omask);
+#endif
 #endif
 		return;
 	}
@@ -1646,9 +1728,13 @@ void logmsg(pri, msg, from, flags)
 			}
 		}
 	}
+#ifdef __gnu_linux__
+	sigprocmask(SIG_UNBLOCK, &mask, NULL);
+#else
 #ifndef SYSV
 	(void) sigsetmask(omask);
 #endif
+#endif
 }
 #if FALSE
 } /* balance parentheses for emacs */
@@ -1771,7 +1857,7 @@ void fprintlog(f, from, flags, msg)
 			dprintf("Not sending message to remote.\n");
 		else {
 			f->f_time = now;
-			(void) snprintf(line, sizeof(line), "<%d>%s\n", f->f_prevpri, \
+			(void) snprintf(line, sizeof(line), "<%d>%s", f->f_prevpri, \
 				(char *) iov[4].iov_base);
 			l = strlen(line);
 			if (l > MAXLINE)
@@ -1815,7 +1901,7 @@ void fprintlog(f, from, flags, msg)
 			v->iov_len = 1;
 		}
 	again:
-		/* f->f_file == -1 is an indicator that the we couldn't
+		/* f->f_file == -1 is an indicator that we couldn't
 		   open the file at startup. */
 		if (f->f_file == -1)
 			break;
@@ -1823,11 +1909,16 @@ void fprintlog(f, from, flags, msg)
 		if (writev(f->f_file, iov, 6) < 0) {
 			int e = errno;
 
-			/* If a named pipe is full, just ignore it for now
-			   - mrn 24 May 96 */
+			/* If a named pipe is full, just ignore it for now */
 			if (f->f_type == F_PIPE && e == EAGAIN)
 				break;
 
+			/* If the filesystem is filled up, just ignore
+			   it for now and continue writing when
+			   possible */
+			if (f->f_type == F_FILE && e == ENOSPC)
+				break;
+
 			(void) close(f->f_file);
 			/*
 			 * Check for EBADF on TTY's due to vhangup() XXX
@@ -1852,7 +1943,7 @@ void fprintlog(f, from, flags, msg)
 				errno = e;
 				logerror(f->f_un.f_fname);
 			}
-		} else if (f->f_flags & SYNC_FILE)
+		} else if (f->f_type == F_FILE && (f->f_flags & SYNC_FILE))
 			(void) fsync(f->f_file);
 		break;
 
@@ -1891,7 +1982,7 @@ void wallmsg(f, iov)
 	register struct filed *f;
 	struct iovec *iov;
 {
-	char p[6 + UNAMESZ];
+	char p[sizeof (_PATH_DEV) + UNAMESZ];
 	register int i;
 	int ttyf, len;
 	static int reenter = 0;
@@ -1899,6 +1990,8 @@ void wallmsg(f, iov)
 	struct utmp *uptr;
 	char greetings[200];
 
+	(void) &len;
+
 	if (reenter++)
 		return;
 
@@ -1913,7 +2006,6 @@ void wallmsg(f, iov)
 	if (fork() == 0) {
 		(void) signal(SIGTERM, SIG_DFL);
 		(void) alarm(0);
-		(void) signal(SIGALRM, endtty);
 #ifndef SYSV
 		(void) signal(SIGTTOU, SIG_IGN);
 		(void) sigsetmask(0);
@@ -1929,7 +2021,7 @@ void wallmsg(f, iov)
 			/* is this slot used? */
 			if (ut.ut_name[0] == '\0')
 				continue;
-			if (ut.ut_type == LOGIN_PROCESS)
+			if (ut.ut_type != USER_PROCESS)
 			        continue;
 			if (!(strcmp (ut.ut_name,"LOGIN"))) /* paranoia */
 			        continue;
@@ -1959,6 +2051,7 @@ void wallmsg(f, iov)
 				iov[1].iov_len = 0;
 			}
 			if (setjmp(ttybuf) == 0) {
+				(void) signal(SIGALRM, endtty);
 				(void) alarm(15);
 				/* open the terminal */
 				ttyf = open(p, O_WRONLY|O_NOCTTY);
@@ -2073,7 +2166,7 @@ void domark()
 	now = time(0);
 	MarkSeq += TIMERINTVL;
 	if (MarkSeq >= MarkInterval) {
-		logmsg(LOG_INFO, "-- MARK --", LocalHostName, ADDDATE|MARK);
+		logmsg(LOG_MARK|LOG_INFO, "-- MARK --", LocalHostName, ADDDATE|MARK);
 		MarkSeq = 0;
 	}
 
@@ -2207,11 +2300,15 @@ void init()
 
 	sp = getservbyname("syslog", "udp");
 	if (sp == NULL) {
+		if (errno == ENOENT) {
+			errno = 0;
+			logerror("The file /etc/services does not seem exist.");
+		}
 		errno = 0;
 		logerror("network logging disabled (syslog/udp service unknown).");
 		logerror("see syslogd(8) for details of whether and how to enable it.");
-		return;
-	}
+		LogPort = 0;
+	} else
 	LogPort = sp->s_port;
 
 	/*
@@ -2605,6 +2702,12 @@ void cfline(line, f)
 	{
 	case '@':
 #ifdef SYSLOG_INET
+		if (!LogPort) {
+			f->f_type = F_UNUSED;
+			logerror("Forward rule without networking enabled");
+			break;
+		}
+
 		(void) strcpy(f->f_un.f_forw.f_hname, ++p);
 		dprintf("forwarding host: %s\n", p);	/*ASP*/
 		if ( (hp = gethostbyname(p)) == NULL ) {
@@ -2637,10 +2740,10 @@ void cfline(line, f)
 		if (syncfile)
 			f->f_flags |= SYNC_FILE;
 		if ( *p == '|' ) {
-			f->f_file = open(++p, O_RDWR|O_NONBLOCK);
+			f->f_file = open(++p, O_RDWR|O_NONBLOCK|O_NOCTTY);
 			f->f_type = F_PIPE;
-	        } else {
-			f->f_file = open(p, O_WRONLY|O_APPEND|O_CREAT|O_NOCTTY,
+		} else {
+			f->f_file = open(p, O_WRONLY|O_APPEND|O_CREAT|O_NONBLOCK|O_NOCTTY,
 					 0644);
 			f->f_type = F_FILE;
 		}
